######################################################################
#
# File: test/unit/v0/apiver/__init__.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

# configured by pytest using `--api` option
# check test/unit/conftest.py:pytest_configure for details
